﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Models.Exercises;
using ClientServer.Models.Exercises.AfterSolutions;
using ClientServer.Models.Exercises.Release;
using ClientServer.Models.Exercises.Tests;
using ClientServer.Models.Users;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;
using ClientServer.Models.Exercises.Solution;
using ClientServer.Models.Files;
using Newtonsoft.Json;

namespace ClientServer.Controllers.Core.Exercises
{
    [Route(Constants.ApiPrefix + "exercise/editor")]
    public class ExerciseEditorController : ControllerWithDb
    {
        public ExerciseEditorController(YapexDbContext context) : base(context)
        {
        }

        private async Task<ExerciseFromBackend> CreateFrontendExercise(Exercise oldExercise, int numReleases,
            bool includeAssetFileData)
        {
            Dictionary<int, List<byte>> markdownFilesDict = null;

            if (includeAssetFileData)
            {
                var markdownFilesIdsToRead = oldExercise.Description.AssetReferences
                    .Select(p => p.FileReferenceMarkdownAssetId)
                    .ToArray();

                var markdownAssetsBasePath = Files.GetUploadFilePath(UploadDirType.MarkdownAssets);

                markdownFilesDict = await Files.ReadUploadedFiles(markdownAssetsBasePath, markdownFilesIdsToRead);
            }


            var exerciseForFrontend = new ExerciseFromBackend()
            {
                Id = oldExercise.Id,
                CreatorId = oldExercise.UserId,
                NumReleases = numReleases,
                Note = oldExercise.Note,
                ExerciseProperties = new ExercisePropertiesFullBase()
                {
                    DisplayName = oldExercise.DisplayName,
                    CanUserCreateFiles = oldExercise.CanUserCreateFiles,
                    CanUserCreateCustomTests = oldExercise.CanUserCreateCustomTests,
                    IsOnlyVisibleToMe = oldExercise.IsOnlyVisibleToOwner,
                    ShortDescription = oldExercise.ShortDescription,
                    UserGroupId = oldExercise.UserGroupId,
                    IsPermanentlyLocked = oldExercise.IsPermanentlyLocked,
                },
                ExerciseDescription = new ExerciseDescriptionFullBase()
                {
                    Id = oldExercise.Description.Id,
                    Content = oldExercise.Description.Content,
                    Assets = oldExercise.Description.AssetReferences.Select(p => new FileWithData()
                    {
                        Id = p.FileReferenceMarkdownAsset.Id,
                        DisplayName = p.FileReferenceMarkdownAsset.OriginalName,
                        MimeType = p.FileReferenceMarkdownAsset.MimeType,
                        Content = (includeAssetFileData && markdownFilesDict != null)
                            ? markdownFilesDict[p.FileReferenceMarkdownAssetId]
                            : new List<byte>()
                    }).ToList()
                },
                MetaData = new ExerciseMetaDataFullBase()
                {
                    Id = oldExercise.MetaData.Id,
                    TagIds = oldExercise.MetaData.TagWithMetaDatas.Select(p => p.TagId).ToList()
                },
                DefaultTestSettings = new TestSettingsFullBase()
                {
                    Id = oldExercise.DefaultCustomTestSettings.Id,
                    MemoryLimitInKb = oldExercise.DefaultCustomTestSettings.MemoryLimitInKb,
                    TimeoutInMs = oldExercise.DefaultCustomTestSettings.TimeoutInMs,
                    MaxDiskSpaceInKb = oldExercise.DefaultCustomTestSettings.MaxDiskSpaceInKb,
                    CompileTimeoutInMs = oldExercise.DefaultCustomTestSettings.CompileTimeoutInMs,
                    CompilerOptions = oldExercise.DefaultCustomTestSettings.CompilerOptions,
                },
                Tests = oldExercise.Tests.Select(p => new ExerciseTestFromBackend()
                {
                    Id = p.Id,
                    DisplayName = p.DisplayName,
                    Content = p.Content,
                    IsSubmitTest = p.IsSubmitTest,
                    Weight = p.Weight,
                    TestTypeId = p.TestTypeId,
                    DisplayIndex = p.DisplayIndex,
                    TestSettings = new TestSettingsFullBase()
                    {
                        Id = p.TestSettings.Id,
                        MemoryLimitInKb = p.TestSettings.MemoryLimitInKb,
                        MaxDiskSpaceInKb = p.TestSettings.MaxDiskSpaceInKb,
                        TimeoutInMs = p.TestSettings.TimeoutInMs,
                        CompileTimeoutInMs = p.TestSettings.CompileTimeoutInMs,
                        CompilerOptions = p.TestSettings.CompilerOptions
                    },
                    Files = p.AssetReferences.Select(asset => new FilePreviewFromBackend()
                    {
                        Id = asset.FileReferenceTestAsset.Id,
                        OriginalName = asset.FileReferenceTestAsset.OriginalName,
                        MimeType = asset.FileReferenceTestAsset.MimeType,
                        SizeInBytes = asset.FileReferenceTestAsset.SizeInBytes
                    }).ToList()
                }).ToList(),
                CodeTemplates = oldExercise.CodeTemplates.Select(p => new ExerciseCodeTemplateFullBase()
                {
                    Id = p.Id,
                    PLangId = p.PLangId,
                    // ReSharper disable once PossibleInvalidOperationException
                    MainFileId = p.MainFileId.Value,
                    TemplateFiles = p.TemplateFiles.Select(file => new ExerciseCodeTemplateFileFullBase()
                    {
                        Id = file.Id,
                        Content = file.Content,
                        IsEditableByUser = file.IsEditableByUser,
                        FileNameWithExtension = file.FileNameWithExtension,
                        IsContentVisibleForUser = file.IsContentVisibleForUser,
                    }).ToList()
                }).ToList(),
            };

            return exerciseForFrontend;
        }

        private async Task<ExerciseFromBackendWithData> CreateFrontendExerciseWithTestFilesData(Exercise oldExercise,
            int numReleases)
        {
            var markdownFilesIdsToRead = oldExercise.Description.AssetReferences
                .Select(p => p.FileReferenceMarkdownAssetId)
                .ToArray();

            var markdownAssetsBasePath = Files.GetUploadFilePath(UploadDirType.MarkdownAssets);

            var markdownFilesDict = await Files.ReadUploadedFiles(markdownAssetsBasePath, markdownFilesIdsToRead);


            var testFilesIdsToRead = oldExercise.Tests.SelectMany(p => p.AssetReferences)
                .Select(p => p.FileReferenceTestAssetId)
                .ToArray();

            var testAssetBasePath = Files.GetUploadFilePath(UploadDirType.TestAssets);

            var testFilesDict = await Files.ReadUploadedFiles(testAssetBasePath, testFilesIdsToRead);

            var exerciseForFrontend = new ExerciseFromBackendWithData()
            {
                Id = oldExercise.Id,
                CreatorId = oldExercise.UserId,
                NumReleases = numReleases,
                Note = oldExercise.Note,
                ExerciseProperties = new ExercisePropertiesFullBase()
                {
                    DisplayName = oldExercise.DisplayName,
                    CanUserCreateFiles = oldExercise.CanUserCreateFiles,
                    CanUserCreateCustomTests = oldExercise.CanUserCreateCustomTests,
                    IsOnlyVisibleToMe = oldExercise.IsOnlyVisibleToOwner,
                    ShortDescription = oldExercise.ShortDescription,
                    UserGroupId = oldExercise.UserGroupId,
                    IsPermanentlyLocked = oldExercise.IsPermanentlyLocked,
                },
                ExerciseDescription = new ExerciseDescriptionFullBase()
                {
                    Id = oldExercise.Description.Id,
                    Content = oldExercise.Description.Content,
                    Assets = oldExercise.Description.AssetReferences.Select(p => new FileWithData()
                    {
                        Id = p.FileReferenceMarkdownAsset.Id,
                        DisplayName = p.FileReferenceMarkdownAsset.OriginalName,
                        MimeType = p.FileReferenceMarkdownAsset.MimeType,
                        Content = markdownFilesDict[p.FileReferenceMarkdownAssetId]
                    }).ToList()
                },
                MetaData = new ExerciseMetaDataFullBase()
                {
                    Id = oldExercise.MetaData.Id,
                    TagIds = oldExercise.MetaData.TagWithMetaDatas.Select(p => p.TagId).ToList()
                },
                DefaultTestSettings = new TestSettingsFullBase()
                {
                    Id = oldExercise.DefaultCustomTestSettings.Id,
                    MemoryLimitInKb = oldExercise.DefaultCustomTestSettings.MemoryLimitInKb,
                    TimeoutInMs = oldExercise.DefaultCustomTestSettings.TimeoutInMs,
                    MaxDiskSpaceInKb = oldExercise.DefaultCustomTestSettings.MaxDiskSpaceInKb,
                    CompileTimeoutInMs = oldExercise.DefaultCustomTestSettings.CompileTimeoutInMs,
                    CompilerOptions = oldExercise.DefaultCustomTestSettings.CompilerOptions,
                },
                Tests = oldExercise.Tests.Select(p => new ExerciseTestFromBackendWithData()
                {
                    Id = p.Id,
                    DisplayName = p.DisplayName,
                    Content = p.Content,
                    IsSubmitTest = p.IsSubmitTest,
                    Weight = p.Weight,
                    TestTypeId = p.TestTypeId,
                    DisplayIndex = p.DisplayIndex,
                    TestSettings = new TestSettingsFullBase()
                    {
                        Id = p.TestSettings.Id,
                        MemoryLimitInKb = p.TestSettings.MemoryLimitInKb,
                        MaxDiskSpaceInKb = p.TestSettings.MaxDiskSpaceInKb,
                        TimeoutInMs = p.TestSettings.TimeoutInMs,
                        CompileTimeoutInMs = p.TestSettings.CompileTimeoutInMs,
                        CompilerOptions = p.TestSettings.CompilerOptions,
                    },
                    Files = p.AssetReferences.Select(asset => new TestAssetFromBackendWithData()
                    {
                        Id = asset.FileReferenceTestAsset.Id,
                        DisplayName = asset.FileReferenceTestAsset.OriginalName,
                        MimeType = asset.FileReferenceTestAsset.MimeType,
                        Size = asset.FileReferenceTestAsset.SizeInBytes,
                        Hash = asset.FileReferenceTestAsset.Hash,
                        Content = testFilesDict != null
                            ? testFilesDict[asset.FileReferenceTestAsset.Id]
                            : null,
                    }).ToList()
                }).ToList(),
                CodeTemplates = oldExercise.CodeTemplates.Select(p => new ExerciseCodeTemplateFullBase()
                {
                    Id = p.Id,
                    PLangId = p.PLangId,
                    // ReSharper disable once PossibleInvalidOperationException
                    MainFileId = p.MainFileId.Value,
                    TemplateFiles = p.TemplateFiles.Select(file => new ExerciseCodeTemplateFileFullBase()
                    {
                        Id = file.Id,
                        Content = file.Content,
                        IsEditableByUser = file.IsEditableByUser,
                        FileNameWithExtension = file.FileNameWithExtension,
                        IsContentVisibleForUser = file.IsContentVisibleForUser,
                    }).ToList()
                }).ToList(),
            };

            return exerciseForFrontend;
        }

        /// <summary>
        /// returns the editor exercise
        /// </summary>
        /// <param name="exerciseId">the exercise to return</param>
        /// <param name="shouldLoadTestAssets">true: the test assets include content, false: not usefull for e.g. copying an exercise</param>
        [HttpGet("{exerciseId}/{shouldLoadTestAssets?}")]
        public async Task GetFullExercise(int exerciseId, bool shouldLoadTestAssets = false)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            //the user group that manages the exercise
            var targetUserGroup =
                await _context.Exercises.Where(p => p.Id == exerciseId).Select(p => p.UserGroupId)
                    .FirstOrDefaultAsync();

            if (
                    !await base.HasGroupPermission(targetUserGroup,
                        permission => permission != null && permission.CanChangeExercises))
                //permission open exercise only when cahnge is allowed
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }

            Exercise oldExercise = await _context.Exercises
                .Where(p => p.Id == exerciseId)
                .Include(p => p.User)
                .Include(p => p.Description)
                .ThenInclude(p => p.AssetReferences)
                .ThenInclude(p => p.FileReferenceMarkdownAsset)
                //meta data
                .Include(p => p.MetaData)
                //tags
                .ThenInclude(p => p.TagWithMetaDatas)
                .Include(p => p.CodeTemplates)
                .ThenInclude(p => p.TemplateFiles)

                //default custom test case settings
                .Include(p => p.DefaultCustomTestSettings)

                //test cases
                .Include(p => p.Tests)
                .ThenInclude(p => p.TestSettings)
                .Include(p => p.Tests)
                .ThenInclude(p => p.AssetReferences)
                .ThenInclude(p => p.FileReferenceTestAsset)
                .FirstOrDefaultAsync();

            if (oldExercise == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }

            //check if the exercise is only visible to the owner...
            if (oldExercise.IsOnlyVisibleToOwner)
            {
                //    //check if the requesting user is the owner...
                if (oldExercise.User == null || oldExercise.User.Id != userId)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise is not visible")));
                    return;
                }
            }


            //every code template neeeds a main file
            if (oldExercise.CodeTemplates.Any(p => p.MainFileId == null))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                            "could not load exercise (no main file)")));
                return;
            }

            var numReleases = await _context.ExerciseReleases
                    .CountAsync(p => p.ExerciseId == oldExercise.Id)
                ;

            if (shouldLoadTestAssets)
            {
                var exerciseForFrontendWithData =
                    await CreateFrontendExerciseWithTestFilesData(oldExercise, numReleases);

                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponseWithData<ExerciseFromBackendWithData>(ResponseCode.Ok,
                            "", exerciseForFrontendWithData)));
                return;
            }


            var exerciseForFrontend = await CreateFrontendExercise(oldExercise, numReleases, true);

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<ExerciseFromBackend>(ResponseCode.Ok,
                        "", exerciseForFrontend)));
        }

        /// <summary>
        /// creates a new exercise
        /// </summary>
        /// <param name="exerciseFromFrontend">the exercise to create (asset file contents are empty because not needed!)</param>
        [HttpPost("create")]
        public async Task CreateFullExercise([FromBody] ExerciseForBackend exerciseFromFrontend)
        {
            //negative ids are possible (from frontend if new) --> mainFileIds can be valid

            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            if (exerciseFromFrontend?.ExerciseProperties == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no exercise")));
                return;
            }

            //the user group that WILL manages the exercise (after the change)
            var targetUserGroup =
                await _context.UserGroups
                    .FirstOrDefaultAsync(p => p.Id == exerciseFromFrontend.ExerciseProperties.UserGroupId);

            if (targetUserGroup == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no managing group selected")));
                return;
            }

            if (
                !await base.HasGroupPermission(targetUserGroup.Id,
                    permission => permission != null && permission.CanCreateExercises))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }


            if (exerciseFromFrontend.CodeTemplates.Count == 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "at least one code template is needed")));
                //TODO MAYBE only currenlty (maybe further exercises does not include a code template and the user has to write every line)
                return;
            }

            #region tags

            var metaData = new MetaData();

            foreach (int tagId in exerciseFromFrontend.MetaData.TagIds)
            {
                Tag tag = await _context.Tags.FirstOrDefaultAsync(p => p.Id == tagId);

                if (tag == null)
                {
                    //another user could have deleted this tag... ignore
                }
                else
                {
                    metaData.TagWithMetaDatas.Add(new TagWithMetaData()
                    {
                        TagId = tag.Id,
                        MetaData = metaData
                    });
                }
            }

            #endregion

            #region default test case settings

            var defaultTestCaseSettings = new DefaultCustomTestSettings
            {
                MemoryLimitInKb = exerciseFromFrontend.DefaultTestSettings.MemoryLimitInKb,
                TimeoutInMs = exerciseFromFrontend.DefaultTestSettings.TimeoutInMs,
                MaxDiskSpaceInKb = exerciseFromFrontend.DefaultTestSettings.MaxDiskSpaceInKb,
                CompileTimeoutInMs = exerciseFromFrontend.DefaultTestSettings.CompileTimeoutInMs,
                CompilerOptions = exerciseFromFrontend.DefaultTestSettings.CompilerOptions,
            };

            //var defaultTestCaseSettings = new List<DefaultCustomTestSetting>();

            //foreach (var defaultTestSettingsFromFrontend in exerciseFromFrontend.DefaultTestSettings)
            //{
            //    defaultTestCaseSettings.Add(new DefaultCustomTestSetting()
            //    {
            //        MemoryLimitInKb = defaultTestSettingsFromFrontend.MemoryLimitInKb,
            //        TimeoutInMs = defaultTestSettingsFromFrontend.TimeoutInMs,
            //        MaxDiskSpaceInKb = defaultTestSettingsFromFrontend.MaxDiskSpaceInKb
            //    });
            //}

            #endregion

            #region tests

            var tests = new List<Test>();

            List<TestType> allTestTypes = await _context.TestTypes.ToListAsync();

            foreach (ExerciseTestForBackend frontendTest in exerciseFromFrontend.Tests)
            {
                var testType = allTestTypes.FirstOrDefault(p => p.Id == frontendTest.TestTypeId);

                if (testType == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "test type not found")));
                    return;
                }

                tests.Add(new Test()
                {
                    DisplayName = frontendTest.DisplayName,
                    Content = frontendTest.Content,
                    IsSubmitTest = frontendTest.IsSubmitTest,
                    Weight = frontendTest.Weight,
                    TestTypeId = testType.Id,
                    DisplayIndex = frontendTest.DisplayIndex,
                    TestSettings = new TestSettings()
                    {
                        MemoryLimitInKb = frontendTest.TestSettings.MemoryLimitInKb,
                        TimeoutInMs = frontendTest.TestSettings.TimeoutInMs,
                        MaxDiskSpaceInKb = frontendTest.TestSettings.MaxDiskSpaceInKb,
                        CompileTimeoutInMs = frontendTest.TestSettings.CompileTimeoutInMs,
                        CompilerOptions = frontendTest.TestSettings.CompilerOptions,
                    },
                });
            }

            #endregion

            #region templates

            var templates = new List<CodeTemplate>();

            var mainFiles = new List<Tuple<CodeTemplate, TemplateFile>>();

            //add every code template
            foreach (ExerciseCodeTemplateFullBase codeTemplate in exerciseFromFrontend.CodeTemplates)
            {
                var files = new List<TemplateFile>();

                TemplateFile mainFile = null;

                //add every file from the code template
                foreach (ExerciseCodeTemplateFileFullBase fileTemplateFromFrontend in codeTemplate.TemplateFiles)
                {
                    var file = new TemplateFile()
                    {
                        Content = fileTemplateFromFrontend.Content,
                        FileNameWithExtension = fileTemplateFromFrontend.FileNameWithExtension,
                        IsEditableByUser = fileTemplateFromFrontend.IsEditableByUser,
                        IsContentVisibleForUser = fileTemplateFromFrontend.IsContentVisibleForUser,
                    };

                    files.Add(file);

                    if (fileTemplateFromFrontend.Id == codeTemplate.MainFileId)
                    {
                        mainFile = file;
                    }
                }

                //we need a main file ... else the solution can not be runned
                if (mainFile == null)
                {
                    //no main class --> no way
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no main file found")));
                    return;
                }

                var newCodeTemplate = new CodeTemplate()
                {
                    //TODO MAYBE better? ef cannot save because the files are not yet saved yet and thus the mainClassFile is curcular
                    //try save all files ... try save main class file -> ref to parent (codeTemplate) -> save??
                    //MainClassFile = mainClassFile,
                    PLangId = codeTemplate.PLangId,
                    TemplateFiles = files
                };

                templates.Add(newCodeTemplate);

                mainFiles.Add(new Tuple<CodeTemplate, TemplateFile>(newCodeTemplate, mainFile));
            }


            if (mainFiles.Count != exerciseFromFrontend.CodeTemplates.Count)
            {
                //theoretically we cannot get here because we check the main file for every code template... 
                //but better check twice ;)
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "main file missing")));
                return;
            }

            #endregion

            #region full exercise

            var newExercise = new Exercise()
            {
                UserId = userId,
                UserGroupId = exerciseFromFrontend.ExerciseProperties.UserGroupId,
                DisplayName = exerciseFromFrontend.ExerciseProperties.DisplayName,
                Description = new ExerciseDescription()
                {
                    Content = exerciseFromFrontend.ExerciseDescription.Content,
                },
                IsPermanentlyLocked = false, //this can only be set after initial creation
                MetaData = metaData,
                CanUserCreateFiles = exerciseFromFrontend.ExerciseProperties.CanUserCreateFiles,
                CanUserCreateCustomTests =
                    exerciseFromFrontend.ExerciseProperties.CanUserCreateCustomTests,
                IsOnlyVisibleToOwner = exerciseFromFrontend.ExerciseProperties.IsOnlyVisibleToMe,
                ShortDescription = exerciseFromFrontend.ExerciseProperties.ShortDescription,
                Tests = tests,
                CodeTemplates = templates,
                DefaultCustomTestSettings = defaultTestCaseSettings,
            };

            #endregion

            newExercise.Note = exerciseFromFrontend.Note;

            try
            {
                using (var transaction = _context.Database.BeginTransaction())
                {
                    try
                    {
                        _context.Exercises.Add(newExercise);
                        await _context.SaveChangesAsync();

                        //now set main files...
                        foreach (var mainFile in mainFiles)
                        {
                            mainFile.Item1.MainFile = mainFile.Item2;
                        }

                        await _context.SaveChangesAsync();

                        transaction.Commit();
                    }
                    catch (Exception ex)
                    {
                        transaction.Rollback();

                        await base.HandleDbError(ex);
                        return;
                    }
                }
            }
            catch (Exception e)
            {
                //see https://www.npgsql.org/doc/transactions.html
                await base.HandleDbError(e);
                return;
            }

            //do not send the full exercise back because when saving in frontend we reload the exercise fully (calling the get method)
            //this way we get all created ids and corrected values
            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<int>(ResponseCode.Ok, "exercise created", newExercise.Id)));
        }

        /// <summary>
        /// updates an existing exercise
        ///
        /// the returned exercise omits the asset contents!
        /// </summary>
        /// <param name="exerciseFromFrontend">the exercise to update (asset file contents are empty because not needed!)</param>
        [HttpPost("update")]
        public async Task ChangeFullExercise([FromBody] ExerciseForBackend exerciseFromFrontend)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            if (exerciseFromFrontend?.ExerciseProperties == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no exercise")));
                return;
            }

            if (exerciseFromFrontend.CodeTemplates.Count == 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "at least one code template is needed")));
                return;
            }


            List<TestType> allTestTypes = await _context.TestTypes.ToListAsync();


            Exercise oldExercise = await _context.Exercises
                .Where(p => p.Id == exerciseFromFrontend.Id)
                .Include(p => p.User)
                .Include(p => p.Description)
                .ThenInclude(p => p.AssetReferences)
                .ThenInclude(p => p.FileReferenceMarkdownAsset)
                //meta data
                .Include(p => p.MetaData)
                //tags
                .ThenInclude(p => p.TagWithMetaDatas)

                //code templates
                .Include(p => p.CodeTemplates)
                .ThenInclude(p => p.TemplateFiles)

                //default custom test case settings
                .Include(p => p.DefaultCustomTestSettings)

                //test cases
                .Include(p => p.Tests)
                .ThenInclude(p => p.TestSettings)
                .Include(p => p.Tests)
                .ThenInclude(p => p.AssetReferences)
                .ThenInclude(p => p.FileReferenceTestAsset)
                .FirstOrDefaultAsync();

            if (oldExercise == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return;
            }


            #region check group permission

            var oldUserGroup =
                await _context.UserGroups.FirstOrDefaultAsync(p =>
                    p.Id == oldExercise.UserGroupId);

            if (oldUserGroup == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "old user group not found")));
                return;
            }

            if (
                !await base.HasGroupPermission(oldUserGroup.Id,
                    permission => permission != null && permission.CanChangeExercises))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission in old group")));
                return;
            }

            UserGroup targetUserGroup = oldUserGroup;

            if (exerciseFromFrontend.ExerciseProperties.UserGroupId != oldUserGroup.Id)
            {
                //the user group that WILL manages the exercise (after the change)
                targetUserGroup =
                    await _context.UserGroups.FirstOrDefaultAsync(p =>
                        p.Id == exerciseFromFrontend.ExerciseProperties.UserGroupId);

                if (targetUserGroup == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "new group not found")));
                    return;
                }

                //moving an exercise is like changing...
                if (
                    !await base.HasGroupPermission(targetUserGroup.Id,
                        permission => permission != null && permission.CanChangeExercises))
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission in new group")));
                    return;
                }
            }

            #endregion


            if (oldExercise.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "exercise is permanently locked")));
                return;
            }

            //check if the exercise is only visible to the owner...
            if (oldExercise.IsOnlyVisibleToOwner)
            {
                //check if the requesting user is the owner...
                if (oldExercise.UserId == null || oldExercise.UserId != userId)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise is not visible")));
                    return;
                }
            } //user could be the owner/creator then he always have access


            //this is the only case we need to check
            //others are
            //old true, new true --> ok
            //old false, new false --> ok
            //old true, new false --> ok (always allowed)
            if (oldExercise.IsOnlyVisibleToOwner == false && exerciseFromFrontend.ExerciseProperties.IsOnlyVisibleToMe)
            {
                //switched to more restrictive mode...
                //this is only allowed if there are no releases for this exercise

                var releasesToChange = await _context.ExerciseReleases
                        .CountAsync(p => p.ExerciseId == oldExercise.Id)
                    ;


                if (releasesToChange > 0)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                "IsOnlyVisibleToMe can only be set to true if there are no releases for this exercise")));
                    return;
                }
            }

            List<int> allPLangsIds = await _context.PLangs
                .Select(p => p.Id)
                .ToListAsync();

            #region tags

            if (exerciseFromFrontend.MetaData == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no meta data")));
                return;
            }

            //check remove old tags
            for (int i = 0; i < oldExercise.MetaData.TagWithMetaDatas.Count; i++)
            {
                var tagConnection = oldExercise.MetaData.TagWithMetaDatas[i];
                if (exerciseFromFrontend.MetaData.TagIds.Any(tagId => tagId == tagConnection.TagId))
                {
                    //old tag still exists in new tags list ...
                }
                else
                {
                    oldExercise.MetaData.TagWithMetaDatas.Remove(tagConnection);
                    i--;
                }
            }

            //add new tags
            foreach (var tagId in exerciseFromFrontend.MetaData.TagIds)
            {
                //check if tag exists
                var tag = await _context.Tags.FirstOrDefaultAsync(p => p.Id == tagId);

                if (tag == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "tag not found")));
                    return;
                }

                var oldTagConnection = oldExercise.MetaData.TagWithMetaDatas.FirstOrDefault(
                    p => p.MetaDataId == oldExercise.MetaData.Id && p.TagId == tag.Id);


                if (oldTagConnection != null)
                {
                    //we already have this tag
                }
                else
                {
                    oldExercise.MetaData.TagWithMetaDatas.Add(new TagWithMetaData()
                    {
                        MetaData = oldExercise.MetaData,
                        Tag = tag
                    });
                }
            }

            #endregion

            #region default test case settings

            if (exerciseFromFrontend.DefaultTestSettings == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no default test settings")));
                return;
            }

            oldExercise.DefaultCustomTestSettings.MaxDiskSpaceInKb =
                exerciseFromFrontend.DefaultTestSettings.MaxDiskSpaceInKb;
            oldExercise.DefaultCustomTestSettings.MemoryLimitInKb =
                exerciseFromFrontend.DefaultTestSettings.MemoryLimitInKb;
            oldExercise.DefaultCustomTestSettings.TimeoutInMs = exerciseFromFrontend.DefaultTestSettings.TimeoutInMs;
            oldExercise.DefaultCustomTestSettings.CompilerOptions =
                exerciseFromFrontend.DefaultTestSettings.CompilerOptions;

            #endregion

            #region tests

            //check old to remove
            for (int i = 0; i < oldExercise.Tests.Count; i++)
            {
                var testCase = oldExercise.Tests[i];
                if (exerciseFromFrontend.Tests.Any(p => p.Id == testCase.Id))
                {
                    //old still exists
                }
                else
                {
                    //remove
                    oldExercise.Tests.Remove(testCase);
                    i--;
                }
            }

            //add new or modify old
            foreach (var testCaseFromBackend in exerciseFromFrontend.Tests)
            {
                var oldExerciseTest = oldExercise.Tests.FirstOrDefault(p => p.Id == testCaseFromBackend.Id);

                var testType = allTestTypes.FirstOrDefault(p => p.Id == testCaseFromBackend.TestTypeId);

                if (testType == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "test type not found")));
                    return;
                }


                if (oldExerciseTest != null)
                {
                    //modify old

                    oldExerciseTest.Content = testCaseFromBackend.Content;
                    oldExerciseTest.DisplayName = testCaseFromBackend.DisplayName;
                    oldExerciseTest.IsSubmitTest = testCaseFromBackend.IsSubmitTest;
                    oldExerciseTest.Weight = testCaseFromBackend.Weight;
                    oldExerciseTest.TestSettings.MemoryLimitInKb = testCaseFromBackend.TestSettings.MemoryLimitInKb;
                    oldExerciseTest.TestSettings.MaxDiskSpaceInKb = testCaseFromBackend.TestSettings.MaxDiskSpaceInKb;
                    oldExerciseTest.TestSettings.TimeoutInMs = testCaseFromBackend.TestSettings.TimeoutInMs;
                    oldExerciseTest.TestSettings.CompilerOptions = testCaseFromBackend.TestSettings.CompilerOptions;
                    oldExerciseTest.DisplayIndex = testCaseFromBackend.DisplayIndex;

                    oldExerciseTest.TestType = testType;
                }
                else
                {
                    //check if valid and create new

                    //create new 
                    oldExercise.Tests.Add(new Test()
                    {
                        Content = testCaseFromBackend.Content,
                        DisplayName = testCaseFromBackend.DisplayName,
                        IsSubmitTest = testCaseFromBackend.IsSubmitTest,
                        Weight = testCaseFromBackend.Weight,
                        DisplayIndex = testCaseFromBackend.DisplayIndex,
                        TestSettings = new TestSettings()
                        {
                            MemoryLimitInKb = testCaseFromBackend.TestSettings.MemoryLimitInKb,
                            MaxDiskSpaceInKb = testCaseFromBackend.TestSettings.MaxDiskSpaceInKb,
                            TimeoutInMs = testCaseFromBackend.TestSettings.TimeoutInMs,
                            CompileTimeoutInMs = testCaseFromBackend.TestSettings.CompileTimeoutInMs,
                            CompilerOptions = testCaseFromBackend.TestSettings.CompilerOptions,
                        },
                        TestType = testType,
                    });
                }
            }


            var tests = new List<Test>();

            foreach (ExerciseTestForBackend frontendTest in exerciseFromFrontend.Tests)
            {
                tests.Add(new Test()
                {
                    DisplayName = frontendTest.DisplayName,
                    Content = frontendTest.Content,
                    IsSubmitTest = frontendTest.IsSubmitTest,
                    Weight = frontendTest.Weight,
                    TestTypeId = frontendTest.TestTypeId,
                    DisplayIndex = frontendTest.DisplayIndex,
                    TestSettings = new TestSettings()
                    {
                        MemoryLimitInKb = frontendTest.TestSettings.MemoryLimitInKb,
                        TimeoutInMs = frontendTest.TestSettings.TimeoutInMs,
                        MaxDiskSpaceInKb = frontendTest.TestSettings.MaxDiskSpaceInKb,
                        CompileTimeoutInMs = frontendTest.TestSettings.CompileTimeoutInMs,
                        CompilerOptions = frontendTest.TestSettings.CompilerOptions,
                    },
                });
            }

            #endregion

            #region templates

            bool didSelfHeadlingMainFile = false;

            //removing or adding template files is only allowed when there is no release... see frontend issue #56

            var hasReleases = await _context.ExerciseReleases
                    .AnyAsync(p => p.ExerciseId == oldExercise.Id)
                ;

            var mainClassFiles = new List<Tuple<CodeTemplate, TemplateFile>>();

            var codeTemplatesToRemove = new List<Tuple<Exercise, CodeTemplate>>();

            //check old to remove
            for (int i = 0; i < oldExercise.CodeTemplates.Count; i++)
            {
                var oldCodeTemplate = oldExercise.CodeTemplates[i];
                if (exerciseFromFrontend.CodeTemplates.Any(p => p.Id == oldCodeTemplate.Id))
                {
                    //still there
                }
                else
                {
                    //remove

                    //not allowed because old solutions will get invalid-ish ... not really but we lose template connection which is bad (probably)
                    if (hasReleases)
                    {
                        await
                            Response.WriteAsync(
                                Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                    "exercise has already releases and therefore you cannot remove code templates")));
                        return;
                    }

                    //main file would give circular dependency... set null
                    //set main file to null save then delete entire code template
                    //oldExercise.CodeTemplates.Remove(oldCodeTemplate);
                    oldCodeTemplate.MainFile = null;

                    codeTemplatesToRemove.Add(new Tuple<Exercise, CodeTemplate>(oldExercise, oldCodeTemplate));
                }
            }

            //check new to add or old to change/modify
            foreach (var codeTemplateFromFrontend in exerciseFromFrontend.CodeTemplates)
            {
                var oldCodeTemplate =
                    oldExercise.CodeTemplates.FirstOrDefault(p => p.Id == codeTemplateFromFrontend.Id);


                if (codeTemplatesToRemove.Any(p => p.Item2 == oldCodeTemplate))
                {
                    //we found a template that will be removed .... to nothing here...
                    continue;
                }

                if (oldCodeTemplate != null)
                {
                    //check files

                    #region change code template 

                    //main file changed??
                    bool hasMainFileChanged = oldCodeTemplate.MainFileId.HasValue &&
                                              oldCodeTemplate.MainFileId.Value !=
                                              codeTemplateFromFrontend.MainFileId;


                    if (hasMainFileChanged && hasReleases)
                    {
                        //not allowed
                        await
                            Response.WriteAsync(
                                Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                    "exercise has already releases and therefore you cannot change the main file")));
                        return;
                    }

                    //check old files to remove
                    for (int i = 0; i < oldCodeTemplate.TemplateFiles.Count; i++)
                    {
                        var oldTemplateFile = oldCodeTemplate.TemplateFiles[i];
                        if (codeTemplateFromFrontend.TemplateFiles.Any(p => p.Id == oldTemplateFile.Id))
                        {
                            //still there
                        }
                        else
                        {
                            if (hasReleases)
                            {
                                //not allowed
                                await
                                    Response.WriteAsync(
                                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                            "exercise has already releases and therefore you cannot remove/add template files")));
                                return;
                            }

                            //remove
                            oldCodeTemplate.TemplateFiles.Remove(oldTemplateFile);
                            i--;
                        }
                    }

                    //check new files to add or old to modify
                    bool mainClassFileFound = false;
                    foreach (var templateFileFromFrontend in codeTemplateFromFrontend.TemplateFiles)
                    {
                        var oldTemplateFile =
                            oldCodeTemplate.TemplateFiles.FirstOrDefault(p => p.Id == templateFileFromFrontend.Id);

                        if (oldTemplateFile != null)
                        {
                            //modify old
                            oldTemplateFile.IsEditableByUser = templateFileFromFrontend.IsEditableByUser;
                            oldTemplateFile.FileNameWithExtension = templateFileFromFrontend.FileNameWithExtension;
                            oldTemplateFile.Content = templateFileFromFrontend.Content;
                            oldTemplateFile.IsContentVisibleForUser = templateFileFromFrontend.IsContentVisibleForUser;

                            if (oldTemplateFile.Id == codeTemplateFromFrontend.MainFileId)
                            {
                                //Main file is an old one
                                mainClassFileFound = true;
                                if (hasMainFileChanged)
                                {
                                    mainClassFiles.Add(new Tuple<CodeTemplate, TemplateFile>(oldCodeTemplate,
                                        oldTemplateFile));
                                }
                            }
                        }
                        else
                        {
                            if (hasReleases)
                            {
                                //not allowed
                                await
                                    Response.WriteAsync(
                                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                                            "exercise has already releases and therefore you cannot add/remove template files")));
                                return;
                            }

                            //add new
                            var newTemplateFile = new TemplateFile()
                            {
                                FileNameWithExtension = templateFileFromFrontend.FileNameWithExtension,
                                IsEditableByUser = templateFileFromFrontend.IsEditableByUser,
                                Content = templateFileFromFrontend.Content,
                                IsContentVisibleForUser = templateFileFromFrontend.IsContentVisibleForUser
                            };
                            oldCodeTemplate.TemplateFiles.Add(newTemplateFile);


                            if (codeTemplateFromFrontend.MainFileId == templateFileFromFrontend.Id)
                            {
                                mainClassFileFound = true;

                                if (hasMainFileChanged)
                                {
                                    //main file changed --> add 
                                    mainClassFiles.Add(new Tuple<CodeTemplate, TemplateFile>(oldCodeTemplate,
                                        newTemplateFile));
                                }
                            }
                        }
                    }

                    if (mainClassFileFound == false)
                    {
                        await
                            Response.WriteAsync(
                                Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "main file not found")));
                        return;
                    }

                    #endregion
                }
                else
                {
                    #region create new code template

                    //add new
                    if (!allPLangsIds.Contains(codeTemplateFromFrontend.PLangId))
                    {
                        await
                            Response.WriteAsync(
                                Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "plang not found")));
                        return;
                    }

                    var newCodeTemplate = new CodeTemplate()
                    {
                        PLangId = codeTemplateFromFrontend.PLangId,
                        MainFile = null,
                        TemplateFiles = new List<TemplateFile>()
                    };

                    bool mainClassFileFound = false;
                    foreach (var templateFileFromFrontend in codeTemplateFromFrontend.TemplateFiles)
                    {
                        TemplateFile file = new TemplateFile()
                        {
                            Content = templateFileFromFrontend.Content,
                            FileNameWithExtension = templateFileFromFrontend.FileNameWithExtension,
                            IsEditableByUser = templateFileFromFrontend.IsEditableByUser,
                            IsContentVisibleForUser = templateFileFromFrontend.IsContentVisibleForUser,
                        };

                        newCodeTemplate.TemplateFiles.Add(file);


                        if (templateFileFromFrontend.Id == codeTemplateFromFrontend.MainFileId)
                        {
                            mainClassFiles.Add(new Tuple<CodeTemplate, TemplateFile>(newCodeTemplate, file));
                            mainClassFileFound = true;
                        }
                    }

                    oldExercise.CodeTemplates.Add(newCodeTemplate);

                    if (mainClassFileFound == false)
                    {
                        
                        //because we removed the main file transaction... we do self healing here...
                        //we use the first file as main file...

                        if (newCodeTemplate.TemplateFiles.Count > 0)
                        {
                            didSelfHeadlingMainFile = true;
                            mainClassFiles.Add(new Tuple<CodeTemplate, TemplateFile>(newCodeTemplate, newCodeTemplate.TemplateFiles.First()));
                            mainClassFileFound = true;
                        }
                        else
                        {
                            await
                                Response.WriteAsync(
                                    Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "main file not found")));
                            return;
                        }                        
                    }

                    #endregion
                }
            }

            #endregion

            //assets can only be added after the exercise is created!! (else we cannot add the connection)

            #region exercise description

            oldExercise.Description.Content = exerciseFromFrontend.ExerciseDescription.Content;

            #endregion

            #region exercise

            //IsOnlyVisibleToMe can only be set by owner

            if (oldExercise.UserId == userId)
            {
                oldExercise.IsOnlyVisibleToOwner = exerciseFromFrontend.ExerciseProperties.IsOnlyVisibleToMe;
            }


            oldExercise.CanUserCreateFiles = exerciseFromFrontend.ExerciseProperties.CanUserCreateFiles;
            oldExercise.CanUserCreateCustomTests =
                exerciseFromFrontend.ExerciseProperties.CanUserCreateCustomTests;
            oldExercise.DisplayName = exerciseFromFrontend.ExerciseProperties.DisplayName;
            oldExercise.ShortDescription = exerciseFromFrontend.ExerciseProperties.ShortDescription;
//            oldExercise.IsPermanentlyLocked //cannot be set via update...we have a separate uri for that


            //managing user group could also have changed
            oldExercise.UserGroup = targetUserGroup;
            oldExercise.Note = exerciseFromFrontend.Note;

            oldExercise.LastUpdatedAt = DateTimeHelper.GetUtcNow();

            #endregion

            try
            {
                using (var transaction = _context.Database.BeginTransaction())
                {
                    try
                    {
                        await _context.SaveChangesAsync();

                        Thread.Sleep(2000);
                        
                        //now main files...
                        if (mainClassFiles.Count > 0)
                        {
                            foreach (Tuple<CodeTemplate, TemplateFile> mainClassFile in mainClassFiles)
                            {
                                mainClassFile.Item1.MainFile = mainClassFile.Item2;
                            }

                            await _context.SaveChangesAsync();
                        }

                        //now remove entire code templates (before we just set the main file to null to avoid circular dependency)
                        if (codeTemplatesToRemove.Count > 0)
                        {
                            foreach (var tuple in codeTemplatesToRemove)
                            {
                                tuple.Item1.CodeTemplates.Remove(tuple.Item2);
                            }

                            await _context.SaveChangesAsync();
                        }


                        transaction.Commit();
                    }
                    catch (Exception ex)
                    {
                        transaction.Rollback();

                        await base.HandleDbError(ex);
                        return;
                    }
                }
            }
            catch (Exception e)
            {
                //see https://www.npgsql.org/doc/transactions.html
                await base.HandleDbError(e);
                return;
            }

            //we need to send back the exercise (or at least the new ids) because we need the new ids
            //e.g. we create the exercise then add a new teSt (gets id -60) and save again but the -60 is never updated...

            var numReleases = await _context.ExerciseReleases
                    .CountAsync(p => p.ExerciseId == oldExercise.Id)
                ;

            //assets are not included on update!!
            var exerciseForFrontend = await CreateFrontendExercise(oldExercise, numReleases, false);

            if (didSelfHeadlingMainFile)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponseWithData<ExerciseFromBackend>(ResponseCode.PartialOk,
                            "main file was set to the first template file", exerciseForFrontend)));
                return;
            }
            
            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<ExerciseFromBackend>(ResponseCode.Ok,
                        "", exerciseForFrontend)));
        }

        [HttpPost("lock/{exerciseId}")]
        public async Task LockExercisePermanently(int exerciseId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var exercise = await _context.Exercises
                .FirstOrDefaultAsync(p => p.Id == exerciseId);

            if (exercise == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise was not found")));
                return;
            }

            var targetUserGroup =
                await _context.UserGroups.FirstOrDefaultAsync(p =>
                    p.Id == exercise.UserGroupId);

            if (targetUserGroup == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "managing group not found")));
                return;
            }

            if (exercise.UserId != userId) //owner can always lock
            {
                //moving an exercise is like changing...
                if (
                    !await base.HasGroupPermission(targetUserGroup.Id,
                        permission => permission != null && permission.CanLockExercisesPermanently))
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                    return;
                }
            }

            exercise.IsPermanentlyLocked = true;
            exercise.IsOnlyVisibleToOwner = false;

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
        }


        /// <summary>
        /// copies the given exercise and returns the new exercise id
        /// </summary>
        /// <param name="exerciseId">the exercise to copy</param>
        [HttpPost("copy/{exerciseId}")]
        public async Task CreateExerciseCopy(int exerciseId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            Exercise oldExercise = await _context.Exercises
                .Where(p => p.Id == exerciseId)
                .Include(p => p.User)
                .Include(p => p.Description)
                .ThenInclude(p => p.AssetReferences)
                .ThenInclude(p => p.FileReferenceMarkdownAsset)
                //meta data
                .Include(p => p.MetaData)
                //tags
                .ThenInclude(p => p.TagWithMetaDatas)
                .Include(p => p.CodeTemplates)
                .ThenInclude(p => p.TemplateFiles)

                //default custom test case settings
                .Include(p => p.DefaultCustomTestSettings)

                //test cases
                .Include(p => p.Tests)
                .ThenInclude(p => p.TestSettings)
                .Include(p => p.Tests)
                .ThenInclude(p => p.AssetReferences)
                .ThenInclude(p => p.FileReferenceTestAsset)
                .AsNoTracking()
                .FirstOrDefaultAsync();

            if (oldExercise == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise was not found")));
                return;
            }

            var targetUserGroup =
                await _context.UserGroups
                    .AsNoTracking()
                    .FirstOrDefaultAsync(p =>
                        p.Id == oldExercise.UserGroupId);

            if (targetUserGroup == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "managing group not found")));
                return;
            }

            if (oldExercise.UserId != userId) //owner can always copy
            {
                //moving an exercise is like changing...
                if (
                    !await base.HasGroupPermission(targetUserGroup.Id,
                        permission => permission != null && permission.CanCreateExercises
                                                         && permission.CanChangeExercises)
                ) //we need change permission because we open the exercise to edit afterwards  
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                    return;
                }
            }

            var user = _context.Users.FirstOrDefaultAsync(p => p.Id == userId);

            if (user == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "user not found")));
                return;
            }

            //copy the exercise... the owner is the current user

            #region description

            var newExDescription = new ExerciseDescription()
            {
                Content = oldExercise.Description.Content,
                AssetReferences = new List<ExerciseDescriptionWithFileAsAssetReference>()
            };

            //copy exercise description asset references

            foreach (var assetReference in oldExercise.Description.AssetReferences)
            {
                var assetReferenceCopy = new ExerciseDescriptionWithFileAsAssetReference()
                {
                    ExerciseDescription = newExDescription,
                    FileReferenceMarkdownAssetId = assetReference.FileReferenceMarkdownAssetId
                };
                newExDescription.AssetReferences.Add(assetReferenceCopy);
            }

            #endregion

            #region code templates

            var newCodeTemplates = new List<CodeTemplate>();
            var mainFiles = new List<Tuple<CodeTemplate, TemplateFile>>();

            foreach (var codeTemplate in oldExercise.CodeTemplates)
            {
                var newCodeTemplate = new CodeTemplate()
                {
                    PLangId = codeTemplate.PLangId,
                    TemplateFiles = new List<TemplateFile>()
                };

                TemplateFile mainFile = null;

                foreach (var templateFile in codeTemplate.TemplateFiles)
                {
                    var newTemplateFile = new TemplateFile()
                    {
                        Content = templateFile.Content,
                        DisplayIndex = templateFile.DisplayIndex,
                        FileNameWithExtension = templateFile.FileNameWithExtension,
                        IsEditableByUser = templateFile.IsEditableByUser,
                        IsContentVisibleForUser = templateFile.IsContentVisibleForUser,
                        SolutionParts = new List<SolutionFile>() //we have no releases --> no solutions
                    };

                    if (templateFile.Id == codeTemplate.MainFileId)
                    {
                        mainFile = newTemplateFile;
                    }

                    newCodeTemplate.TemplateFiles.Add(newTemplateFile);
                }

                if (mainFile == null)
                {
                    //no main class --> no way
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "no main file found")));
                    return;
                }

                mainFiles.Add(new Tuple<CodeTemplate, TemplateFile>(newCodeTemplate, mainFile));
                newCodeTemplates.Add(newCodeTemplate);
            }

            #endregion

            #region tags

            var newMetaData = new MetaData();

            foreach (var tagWithMetaData in oldExercise.MetaData.TagWithMetaDatas)
            {
                var newTag = new TagWithMetaData()
                {
                    TagId = tagWithMetaData.TagId,
                    MetaData = newMetaData
                };
                newMetaData.TagWithMetaDatas.Add(newTag);
            }

            #endregion

            #region tests

            var newTests = new List<Test>();

            foreach (var exerciseTest in oldExercise.Tests)
            {
                var newTest = new Test()
                {
                    Content = exerciseTest.Content,
                    Weight = exerciseTest.Weight,
                    DisplayIndex = exerciseTest.DisplayIndex,
                    DisplayName = exerciseTest.DisplayName,
                    TestTypeId = exerciseTest.TestTypeId,
                    IsSubmitTest = exerciseTest.IsSubmitTest,
                    TestSettings = new TestSettings()
                    {
                        CompileTimeoutInMs = exerciseTest.TestSettings.CompileTimeoutInMs,
                        TimeoutInMs = exerciseTest.TestSettings.TimeoutInMs,
                        MemoryLimitInKb = exerciseTest.TestSettings.MemoryLimitInKb,
                        MaxDiskSpaceInKb = exerciseTest.TestSettings.MaxDiskSpaceInKb,
                        CompilerOptions = exerciseTest.TestSettings.CompilerOptions,
                    },
                    AssetReferences = new List<TestWithFileAsAssetReference>(),
                };

                //copy test asset references

                foreach (var assetReference in exerciseTest.AssetReferences)
                {
                    var newAssetReference = new TestWithFileAsAssetReference
                    {
                        Test = newTest,
                        FileReferenceTestAssetId = assetReference.FileReferenceTestAssetId
                    };

                    newTest.AssetReferences.Add(newAssetReference);
                }

                newTests.Add(newTest);
            }

            #endregion

            var exCopy = new Exercise()
            {
                DisplayName = oldExercise.DisplayName + " - Copy",
                ShortDescription = oldExercise.ShortDescription,
                CanUserCreateCustomTests = oldExercise.CanUserCreateCustomTests,
                CanUserCreateFiles = oldExercise.CanUserCreateFiles,
                IsOnlyVisibleToOwner = oldExercise.IsOnlyVisibleToOwner,
                IsPermanentlyLocked = false, //this would make no sense... if we take it
                Note = oldExercise.Note,
                Releases = new List<ExerciseRelease>(),
                UserId = userId,
                UserGroupId = oldExercise.UserGroupId,
                Description = newExDescription,
                CodeTemplates = newCodeTemplates,
                MetaData = newMetaData,
                Tests = newTests,
                DefaultCustomTestSettings = new DefaultCustomTestSettings()
                {
                    TimeoutInMs = oldExercise.DefaultCustomTestSettings.TimeoutInMs,
                    CompileTimeoutInMs = oldExercise.DefaultCustomTestSettings.CompileTimeoutInMs,
                    MaxDiskSpaceInKb = oldExercise.DefaultCustomTestSettings.MaxDiskSpaceInKb,
                    MemoryLimitInKb = oldExercise.DefaultCustomTestSettings.MemoryLimitInKb,
                    CompilerOptions = oldExercise.DefaultCustomTestSettings.CompilerOptions,
                },
            };

            string jsonTests = "";
            //string jsonTests2 = "";
            try
            {
                
                //NOT WORKING AT ALL in production environment...
                //we already fail at _context.Database.BeginTransaction() (locally works)
                //sometimes we can force this locally when we login as a different user (other browser) save a solution/exercise
                //e.g. via shortcut and try to create a copy (time between the two actions must be < ~0.5s?)
                //without transaction this works fine...
                //in the worst case we have a exercise but no main file!!
                
                //see https://docs.microsoft.com/en-us/previous-versions/sql/sql-server-2008-r2/ms189122(v=sql.105)?redirectedfrom=MSDN
                //also see https://coderwall.com/p/jnniww/why-you-shouldn-t-use-entity-framework-with-transactions
                //default is read committed
                
//                using (var transaction = _context.Database.BeginTransaction(IsolationLevel.Snapshot))
                    
//                    using (var transaction = _context.Database.BeginTransaction(IsolationLevel.Serializable))
//                using (var transaction = _context.Database.BeginTransaction())
//                {
                    try
                    {
                        _context.Exercises.Add(exCopy);

                        //TODO does this help??
                        foreach (var test in newTests)
                        {
                            test.TestSettings.Id = test.Id;
                        }

                        jsonTests = JsonConvert.SerializeObject(exCopy.Tests, Formatting.Indented,
                            new JsonSerializerSettings()
                            {
                                ReferenceLoopHandling = ReferenceLoopHandling.Serialize,
                                PreserveReferencesHandling = PreserveReferencesHandling.All
                            });
                        
                        await _context.SaveChangesAsync();

                        //now set main files...
                        foreach (var mainFile in mainFiles)
                        {
                            mainFile.Item1.MainFile = mainFile.Item2;
                        }

                        await _context.SaveChangesAsync();

//                        transaction.Commit();
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(ex);

                        Console.WriteLine("DBG: " + jsonTests);

//                        transaction.Rollback();


                        await base.HandleDbError(ex);
                        return;
                    }
//                }
            }
            catch (Exception e)
            {
                
                //we get the error here: ERROR:  insert or update on table "TestCaseSettingses" violates foreign key constraint "FK_TestCaseSettingses_Tests_TestId
                //DETAIL:  Key (TestId)=(0) is not present in table "Tests".
//                Console.WriteLine("DBG: " + jsonTests);
                
                //see https://www.npgsql.org/doc/transactions.html
                await base.HandleDbError(e);
                return;
            }

            //do not send the full exercise back because when saving in frontend we reload the exercise fully (calling the get method)
            //this way we get all created ids and corrected values
            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<int>(ResponseCode.Ok, "", exCopy.Id)));
        }

        /// <summary>
        /// checks if the exercise is valid and if the user has access to modify it
        /// </summary>
        /// <param name="oldExercise"></param>
        /// <param name="userId"></param>
        /// <param name="checkDescription"></param>
        /// <returns></returns>
        private async Task<bool> _HasPermissionAndIsValid(Exercise oldExercise, int userId, bool checkDescription)
        {
            if (oldExercise == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise not found")));
                return false;
            }

            if (checkDescription && oldExercise.Description == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise description not found")));
                return false;
            }


            #region check group permission

            var oldUserGroup =
                await _context.UserGroups.FirstOrDefaultAsync(p =>
                    p.Id == oldExercise.UserGroupId);

            if (oldUserGroup == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "old user group not found")));
                return false;
            }

            if (
                !await base.HasGroupPermission(oldUserGroup.Id,
                    permission => permission != null && permission.CanChangeExercises))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission in old group")));
                return false;
            }

            #endregion


            if (oldExercise.IsPermanentlyLocked)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "exercise is permanently locked")));
                return false;
            }

            //check if the exercise is only visible to the owner...
            if (oldExercise.IsOnlyVisibleToOwner)
            {
                //check if the requesting user is the owner...
                if (oldExercise.UserId == null || oldExercise.UserId != userId)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "exercise is not visible")));
                    return false;
                }
            } //user could be the owner/creator then he always have access

            return true;
        }

        #region exercise description asset files handling

        /// <summary>
        /// this will upload (receive) the file and connect it to the exercise description
        /// </summary>
        /// <param name="exerciseId"></param>
        /// <param name="file"></param>
        /// <returns></returns>
        [HttpPost("upload/create/description/{exerciseId}")]
        public async Task UploadExerciseDescriptionAssetFile(int exerciseId, IFormFile file)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            Exercise oldExercise = await _context.Exercises
                .Where(p => p.Id == exerciseId)
                .Include(p => p.Description)
                .FirstOrDefaultAsync();

            if (await _HasPermissionAndIsValid(oldExercise, userId, true) == false) return;


            if (file == null || file.Length == 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no files selected")));
                return;
            }

            var targetHardDrivePath = Files.GetUploadFilePath(UploadDirType.MarkdownAssets);
            FileReferenceMarkdownAsset fileRef = null;
            try
            {
                using (var transaction = await _context.Database.BeginTransactionAsync())
                {
                    try
                    {
                        fileRef = new FileReferenceMarkdownAsset()
                        {
                            Hash = "",
                            MimeType = file.ContentType,
                            OriginalName = file.FileName,
                        };

                        var connection = new ExerciseDescriptionWithFileAsAssetReference()
                        {
                            FileReferenceMarkdownAsset = fileRef,
                            ExerciseDescription = oldExercise.Description
                        };

                        _context.ExerciseDescriptionWithFileAsAssetReferences.Add(connection);
                        await _context.SaveChangesAsync();

                        Tuple<string, long> _tuple =
                            await Files.CreateUploadedFile(targetHardDrivePath, fileRef.Id, file);
                        fileRef.Hash = _tuple.Item1;
                        fileRef.SizeInBytes = _tuple.Item2;

                        await _context.SaveChangesAsync();
                        transaction.Commit();
                    }
                    catch (Exception e)
                    {
                        transaction.Rollback();
                        await base.HandleDbError(e);
                        return;
                    }
                }
            }
            catch (Exception e)
            {
                await base.HandleDbError(e);
                return;
            }


            var assetPreviewForFrontend = new FilePreviewFromBackend()
            {
                Id = fileRef.Id,
                MimeType = fileRef.MimeType,
                OriginalName = fileRef.OriginalName,
                SizeInBytes = fileRef.SizeInBytes
            };


            await
                Response.WriteAsync(
                    Jc.Serialize(
                        new BasicResponseWithData<FilePreviewFromBackend>(ResponseCode.Ok, "",
                            assetPreviewForFrontend)));
        }

        /// <summary>
        /// adds an already uploaded file to the exercise description
        /// </summary>
        /// <param name="exerciseId"></param>
        /// <param name="fileReferenceId"></param>
        /// <returns></returns>
        [HttpPost("upload/attach/description/{exerciseId}/{fileReferenceId}")]
        public async Task AttachExitingFileToExerciseDescription(int exerciseId, int fileReferenceId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            Exercise oldExercise = await _context.Exercises
                .Where(p => p.Id == exerciseId)
                .Include(p => p.Description)
                .FirstOrDefaultAsync();


            if (await _HasPermissionAndIsValid(oldExercise, userId, true) == false) return;


            var fileRef = await _context.FileReferenceMarkdownAssets
                .FirstOrDefaultAsync(p => p.Id == fileReferenceId);

            if (fileRef == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "asset file not found")));
                return;
            }

            var connection = new ExerciseDescriptionWithFileAsAssetReference()
            {
                FileReferenceMarkdownAsset = fileRef,
                ExerciseDescriptionId = oldExercise.Description.Id
            };

            try
            {
                _context.ExerciseDescriptionWithFileAsAssetReferences.Add(connection);
                await _context.SaveChangesAsync();
            }
            catch (Exception e)
            {
                await base.HandleDbError(e);
                return;
            }

            //we don't need to send back information because the frontend need to load the file anyway and will
            //call GetSingleExerciseDescriptionAssetFile which includes all needed information

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
        }


        ///  <summary>
        ///  removes the file from the exercise (exercise description)
        /// 
        ///  IF the file was only attached to this last exercise then the file gets deleted
        ///    (because there are no other connection to this file)
        ///  </summary>
        ///  <param name="exerciseId"></param>
        /// <param name="fileReferenceId"></param>
        /// <returns></returns>
        [HttpDelete("upload/detach/description/{exerciseId}/{fileReferenceId}")]
        public async Task DetachOrDeleteFileFromExerciseDescription(int exerciseId, int fileReferenceId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            Exercise oldExercise = await _context.Exercises
                .Where(p => p.Id == exerciseId)
                .Include(p => p.Description)
                .FirstOrDefaultAsync();

            if (await _HasPermissionAndIsValid(oldExercise, userId, true) == false) return;


            var assetFileReference = await _context.ExerciseDescriptionWithFileAsAssetReferences
                .FirstOrDefaultAsync(p =>
                    p.FileReferenceMarkdownAssetId == fileReferenceId &&
                    p.ExerciseDescriptionId == oldExercise.Description.Id);

            if (assetFileReference == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "file reference connection not found")));
                return;
            }

            //then check if this was the last reference to this file from an exercise...

            var referenceCount = await _context.ExerciseDescriptionWithFileAsAssetReferences
                    .CountAsync(p => p.FileReferenceMarkdownAssetId == assetFileReference.FileReferenceMarkdownAssetId)
                ;

            if (referenceCount == 1) //this is the last reference
            {
                //no exercise has a reference to this file... delete


                var fileReference = await _context.FileReferenceMarkdownAssets
                    .FirstOrDefaultAsync(p => p.Id == assetFileReference.FileReferenceMarkdownAssetId);

                if (fileReference == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "file reference not found")));
                    return;
                }


                using (var transaction = await _context.Database.BeginTransactionAsync())
                {
                    try
                    {
                        _context.FileReferenceMarkdownAssets.Remove(fileReference);
                        await _context.SaveChangesAsync();

                        var targetHardDrivePath = Files.GetUploadFilePath(UploadDirType.MarkdownAssets);

                        Files.DeleteUploadedFile(targetHardDrivePath, fileReference.Id);

                        transaction.Commit();
                    }
                    catch (Exception e)
                    {
                        transaction.Rollback();
                        await base.HandleDbError(e);
                        return;
                    }
                }
            }
            else
            {
                try
                {
                    _context.ExerciseDescriptionWithFileAsAssetReferences.Remove(assetFileReference);
                    await _context.SaveChangesAsync();
                }
                catch (Exception e)
                {
                    await base.HandleDbError(e);
                    return;
                }
            }


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
        }

        /// <summary>
        /// returns all possible uploaded file ids the user can use TODO
        /// </summary>
        /// <returns></returns>
        [HttpGet("upload/get/description/all")]
        public async Task GetPossibleExerciseDescriptionFiles()
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            //get all group (ids) where the user as permission to access the exercise


            var groupIdsWhereUserHasAccess = await _context.UserWithUserGroups
                    .Where(p => p.UserId == userId &&
                                (p.GroupRole.GroupRolePermission.CanChangeExercises))
                    .Select(p => p.UserGroupId)
                    .ToListAsync()
                ;

            //get files from exercises where the user is in the managing group 

            //TODO this can fail if the group id list gets too long
            var references = await _context.ExerciseDescriptionWithFileAsAssetReferences
                    .Include(p => p.FileReferenceMarkdownAsset)
                    .Where(p => p.ExerciseDescription.Exercise.IsOnlyVisibleToOwner == false
                                && groupIdsWhereUserHasAccess.Contains(p.ExerciseDescription.Exercise.UserGroupId))
                    .ToListAsync()
                ;


            var allFileReferencesForFrontend = references
                .DistinctBy(p =>
                    p.FileReferenceMarkdownAssetId) //a file can be connected to two exercises in the same group
                .Select(p => new FilePreviewFromBackend()
                {
                    Id = p.FileReferenceMarkdownAsset.Id,
                    SizeInBytes = p.FileReferenceMarkdownAsset.SizeInBytes,
                    OriginalName = p.FileReferenceMarkdownAsset.OriginalName,
                    MimeType = p.FileReferenceMarkdownAsset.MimeType
                }).ToList();

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<List<FilePreviewFromBackend>>(ResponseCode.Ok, "",
                        allFileReferencesForFrontend)));
        }


        /// <summary>
        /// this is needed when we attach an existing file to an exercise then we need to content in the frontend
        /// </summary>
        /// <param name="exerciseId"></param>
        /// <param name="fileReferenceId"></param>
        /// <returns></returns>
        [HttpGet("upload/get/description/content/{exerciseId}/{fileReferenceId}")]
        public async Task GetSingleExerciseDescriptionAssetFile(int exerciseId, int fileReferenceId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            Exercise oldExercise = await _context.Exercises
                .Where(p => p.Id == exerciseId)
                .Include(p => p.Description)
                .FirstOrDefaultAsync();


            if (await _HasPermissionAndIsValid(oldExercise, userId, true) == false) return;


            //file should only be read if it is connected to the exercise

            var connection = await _context.ExerciseDescriptionWithFileAsAssetReferences
                    .Where(p => p.ExerciseDescriptionId == oldExercise.Description.Id &&
                                p.FileReferenceMarkdownAssetId == fileReferenceId)
                    .AsNoTracking()
                    .FirstOrDefaultAsync()
                ;

            if (connection == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound,
                            "file is not connected to the exercise")));
                return;
            }

            var fileRef = await _context.FileReferenceMarkdownAssets.FirstOrDefaultAsync(p => p.Id == fileReferenceId);

            if (fileRef == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "file was not found")));
                return;
            }

            var basePath = Files.GetUploadFilePath(UploadDirType.MarkdownAssets);

            var filesDict = await Files.ReadUploadedFiles(basePath, fileReferenceId);


            var assetForFrontend = new FileWithData()
            {
                Id = fileRef.Id,
                DisplayName = fileRef.OriginalName,
                MimeType = fileRef.MimeType,
                Content = filesDict[fileReferenceId]
            };

            await
                Response.WriteAsync(
                    Jc.Serialize(
                        new BasicResponseWithData<FileWithData>(ResponseCode.Ok, "", assetForFrontend)));
        }

        #endregion


        #region exercise test asset files handling

        /// <summary>
        /// this will upload (receive) the file and connect it to the test
        /// </summary>
        /// <param name="exerciseId"></param>
        /// <param name="testId"></param>
        /// <param name="file"></param>
        /// <returns></returns>
        [HttpPost("upload/create/tests/{exerciseId}/{testId}")]
        public async Task UploadExerciseTestAssetFile(int exerciseId, int testId, IFormFile file)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            Exercise oldExercise = await _context.Exercises
                .Where(p => p.Id == exerciseId)
                .FirstOrDefaultAsync();

            if (await _HasPermissionAndIsValid(oldExercise, userId, false) == false) return;


            var oldTest = await _context.Tests
                .FirstOrDefaultAsync(p => p.Id == testId && p.ExerciseId == exerciseId);

            if (oldTest == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "test not found")));
                return;
            }

            if (file == null || file.Length == 0)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no files selected")));
                return;
            }

            var targetHardDrivePath = Files.GetUploadFilePath(UploadDirType.TestAssets);
            FileReferenceTestAsset fileRef = null;
            try
            {
                using (var transaction = await _context.Database.BeginTransactionAsync())
                {
                    try
                    {
                        fileRef = new FileReferenceTestAsset()
                        {
                            Hash = "",
                            MimeType = file.ContentType,
                            OriginalName = file.FileName,
                        };

                        var connection = new TestWithFileAsAssetReference()
                        {
                            FileReferenceTestAsset = fileRef,
                            Test = oldTest
                        };

                        _context.TestWithFileAsAssetReferences.Add(connection);
                        await _context.SaveChangesAsync();

                        Tuple<string, long> _tuple =
                            await Files.CreateUploadedFile(targetHardDrivePath, fileRef.Id, file);
                        fileRef.Hash = _tuple.Item1;
                        fileRef.SizeInBytes = _tuple.Item2;

                        await _context.SaveChangesAsync();
                        transaction.Commit();
                    }
                    catch (Exception e)
                    {
                        transaction.Rollback();
                        await base.HandleDbError(e);
                        return;
                    }
                }
            }
            catch (Exception e)
            {
                await base.HandleDbError(e);
                return;
            }


            var assetPreviewForFrontend = new FilePreviewFromBackend()
            {
                Id = fileRef.Id,
                MimeType = fileRef.MimeType,
                OriginalName = fileRef.OriginalName,
                SizeInBytes = fileRef.SizeInBytes
            };

            await
                Response.WriteAsync(
                    Jc.Serialize(
                        new BasicResponseWithData<FilePreviewFromBackend>(ResponseCode.Ok, "",
                            assetPreviewForFrontend)));
        }


        /// <summary>
        /// adds an already uploaded file to the test
        /// </summary>
        /// <param name="exerciseId"></param>
        /// <param name="testId"></param>
        /// <param name="fileReferenceId"></param>
        /// <returns></returns>
        [HttpPost("upload/attach/tests/{exerciseId}/{testId}/{fileReferenceId}")]
        public async Task AttachExitingFileToExerciseTest(int exerciseId, int testId, int fileReferenceId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            Exercise oldExercise = await _context.Exercises
                .Where(p => p.Id == exerciseId)
                .FirstOrDefaultAsync();

            if (await _HasPermissionAndIsValid(oldExercise, userId, false) == false) return;

            var oldTest = await _context.Tests
                .FirstOrDefaultAsync(p => p.Id == testId && p.ExerciseId == exerciseId);

            if (oldTest == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "test not found")));
                return;
            }

            var fileRef = await _context.FileReferenceTestAssets
                .FirstOrDefaultAsync(p => p.Id == fileReferenceId);

            if (fileRef == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "asset file not found")));
                return;
            }

            var connection = new TestWithFileAsAssetReference()
            {
                FileReferenceTestAsset = fileRef,
                Test = oldTest
            };

            try
            {
                _context.TestWithFileAsAssetReferences.Add(connection);
                await _context.SaveChangesAsync();
            }
            catch (Exception e)
            {
                await base.HandleDbError(e);
                return;
            }

            //we need to send some information so the frontend can display the asset file

            var filePreview = new FilePreviewFromBackend()
            {
                Id = fileRef.Id,
                OriginalName = fileRef.OriginalName,
                SizeInBytes = fileRef.SizeInBytes,
                MimeType = fileRef.MimeType
            };


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<FilePreviewFromBackend>(ResponseCode.Ok, "", filePreview)));
        }

        ///  <summary>
        ///  removes the file from the test 
        /// 
        ///  IF the file was only attached to this last test then the file gets deleted
        ///    (because there are no other connection to this file)
        ///  </summary>
        ///  <param name="exerciseId"></param>
        /// <param name="testId"></param>
        /// <param name="fileReferenceId"></param>
        /// <returns></returns>
        [HttpDelete("upload/detach/tests/{exerciseId}/{testId}/{fileReferenceId}")]
        public async Task DetachOrDeleteFileFromExerciseTest(int exerciseId, int testId, int fileReferenceId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            Exercise oldExercise = await _context.Exercises
                .Where(p => p.Id == exerciseId)
                .FirstOrDefaultAsync();

            if (await _HasPermissionAndIsValid(oldExercise, userId, false) == false) return;

            var oldTest = await _context.Tests
                .FirstOrDefaultAsync(p => p.Id == testId && p.ExerciseId == exerciseId);

            if (oldTest == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "test not found")));
                return;
            }

            var assetFileReference = await _context.TestWithFileAsAssetReferences.FirstOrDefaultAsync(p =>
                p.FileReferenceTestAssetId == fileReferenceId && p.TestId == oldTest.Id);

            if (assetFileReference == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "file reference connection not found")));
                return;
            }

            //then check if this was the last reference to this file from an exercise...

            var referenceCount = await _context.TestWithFileAsAssetReferences
                    .CountAsync(p => p.FileReferenceTestAssetId == assetFileReference.FileReferenceTestAssetId)
                ;

            if (referenceCount == 1) //this is the last reference
            {
                //no test has a reference to this file... delete


                var fileReference = await _context.FileReferenceTestAssets
                    .FirstOrDefaultAsync(p => p.Id == assetFileReference.FileReferenceTestAssetId);


                if (fileReference == null)
                {
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "file reference not found")));
                    return;
                }

                using (var transaction = await _context.Database.BeginTransactionAsync())
                {
                    try
                    {
                        _context.FileReferenceTestAssets.Remove(fileReference);
                        await _context.SaveChangesAsync();

                        var targetHardDrivePath = Files.GetUploadFilePath(UploadDirType.TestAssets);

                        Files.DeleteUploadedFile(targetHardDrivePath, fileReference.Id);

                        transaction.Commit();
                    }
                    catch (Exception e)
                    {
                        transaction.Rollback();
                        await base.HandleDbError(e);
                        return;
                    }
                }
            }
            else
            {
                try
                {
                    _context.TestWithFileAsAssetReferences.Remove(assetFileReference);
                    await _context.SaveChangesAsync();
                }
                catch (Exception e)
                {
                    await base.HandleDbError(e);
                    return;
                }
            }


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
        }

        /// <summary>
        /// returns all possible uploaded file ids the user can use
        /// </summary>
        /// <returns></returns>
        [HttpGet("upload/get/tests/all")]
        public async Task GetPossibleExerciseTestFiles()
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            var groupIdsWhereUserHasAccess = await _context.UserWithUserGroups
                    .Where(p => p.UserId == userId &&
                                (p.GroupRole.GroupRolePermission.CanChangeExercises))
                    .Select(p => p.UserGroupId)
                    .ToListAsync()
                ;

            //TODO this can fail if the group id list gets too long
            var references = await _context.TestWithFileAsAssetReferences
                    .Include(p => p.FileReferenceTestAsset)
                    .Where(p => p.Test.Exercise.IsOnlyVisibleToOwner == false
                                && groupIdsWhereUserHasAccess.Contains(p.Test.Exercise.UserGroupId))
                    .ToListAsync()
                ;

            var allFileReferencesForFrontend = references
                .DistinctBy(p =>
                    p.FileReferenceTestAssetId) //a test file can be connected to two exercises in the same group
                .Select(p => new FilePreviewFromBackend()
                {
                    Id = p.FileReferenceTestAsset.Id,
                    SizeInBytes = p.FileReferenceTestAsset.SizeInBytes,
                    OriginalName = p.FileReferenceTestAsset.OriginalName,
                    MimeType = p.FileReferenceTestAsset.MimeType
                }).ToList();

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<List<FilePreviewFromBackend>>(ResponseCode.Ok, "",
                        allFileReferencesForFrontend)));
        }

        #endregion
    }

    #region types

    #region full bases

    /// <summary>
    /// the properties for an exercise (in the backend these are part of the exercise but not in the frontend (have own panel)
    /// </summary>
    public class ExercisePropertiesFullBase
    {
        /// <summary>
        /// the display name of the exercise
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the short description
        /// </summary>
        public string ShortDescription { get; set; }

        /// <summary>
        /// true: user can create new files, false: can only modify the given files
        /// </summary>
        public bool CanUserCreateFiles { get; set; }

        /// <summary>
        /// true: user can create custom tests, false: not
        /// </summary>
        public bool CanUserCreateCustomTests { get; set; }

        /// <summary>
        /// true: the exercise is only visible to me and releases can only be accessed by owner
        /// </summary>
        public bool IsOnlyVisibleToMe { get; set; }

        /// <summary>
        /// exercise and all dependent entities should be readonly (no edit/no delete)!
        /// </summary>
        public bool IsPermanentlyLocked { get; set; }

        /// <summary>
        /// the group id where the exercise will belong to
        /// </summary>
        public int UserGroupId { get; set; }
    }

    /// <summary>
    /// a description for an exercise
    /// </summary>
    public class ExerciseDescriptionFullBase
    {
        public int Id { get; set; }

        /// <summary>
        /// the content
        /// </summary>
        public string Content { get; set; }

        /// <summary>
        /// the assets (files) are only send from backend to frontend
        /// frontend need to add these through separate requests!
        ///
        /// FROM frontend to backend this is always empty!
        /// </summary>
        public List<FileWithData> Assets { get; set; }

        public ExerciseDescriptionFullBase()
        {
            Assets = new List<FileWithData>();
        }
    }

    /// <summary>
    /// the exercise asset for the exercise description, because we need the content to display e.g. an image...
    /// </summary>
    public class FileWithData
    {
        public int Id { get; set; }

        /// <summary>
        /// the display name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the mime type
        /// </summary>
        public string MimeType { get; set; }

        /// <summary>
        /// the content
        /// can be null if file could not be read (or does not exist)
        /// </summary>
        public List<byte> Content { get; set; } //backend --> frontend its a list | frontend --> backend its an array...
    }

    /// <summary>
    /// the meta data for the exercise (currently only tags)
    /// </summary>
    public class ExerciseMetaDataFullBase
    {
        public int Id { get; set; }

        /// <summary>
        /// the list of tag ids
        /// </summary>
        public List<int> TagIds { get; set; }

        public ExerciseMetaDataFullBase()
        {
            TagIds = new List<int>();
        }
    }


    /// <summary>
    /// an exercise code template for a plang
    /// </summary>
    public class ExerciseCodeTemplateFullBase
    {
        public int Id { get; set; }

        /// <summary>
        /// the plang this code template is for
        /// </summary>
        public int PLangId { get; set; }

        /// <summary>
        /// the main file
        /// </summary>
        public int MainFileId { get; set; }

        /// <summary>
        /// the code template files
        /// </summary>
        public List<ExerciseCodeTemplateFileFullBase> TemplateFiles { get; set; }

        public ExerciseCodeTemplateFullBase()
        {
            TemplateFiles = new List<ExerciseCodeTemplateFileFullBase>();
        }
    }

    /// <summary>
    /// a code template file, some lines of code the tutor wants to give the students (assistance)
    /// </summary>
    public class ExerciseCodeTemplateFileFullBase : ISolutionFile
    {
        public int Id { get; set; }

        /// <summary>
        /// the file name with extension
        /// </summary>
        public string FileNameWithExtension { get; set; }

        /// <summary>
        /// the order index how the file is displayed in the editor
        /// NOT USED YET
        /// </summary>
        public int DisplayIndex { get; set; }

        /// <summary>
        /// true: user can edit the file, false: not
        /// </summary>
        public bool IsEditableByUser { get; set; }

        /// <summary>
        /// true: file content should be visible by the user,
        /// false: not visible to the user (but user could get all files in the dir...)
        /// </summary>
        public bool IsContentVisibleForUser { get; set; }

        /// <summary>
        /// the content of the file
        /// </summary>
        public string Content { get; set; }
    }


    /// <summary>
    /// test settings for a test
    /// </summary>
    public class TestSettingsFullBase
    {
        public int Id { get; set; }

        /// <summary>
        /// the test timeout in ms
        /// </summary>
        public int TimeoutInMs { get; set; }

        /// <summary>
        /// the compile timeout in ms
        /// </summary>
        public int CompileTimeoutInMs { get; set; }

        /// <summary>
        /// the memory limit in kb
        /// </summary>
        public int MemoryLimitInKb { get; set; }

        /// <summary>
        /// the max disk space to users program can write to
        /// </summary>
        public int MaxDiskSpaceInKb { get; set; }

        /// <summary>
        /// some compiler options .g. -Xlint options needs to be separated by whitespaces (like normal command args)
        /// </summary>
        public string CompilerOptions { get; set; }
    }

    #endregion

    #region different types for front and backend

    #region exercise

    /// <summary>
    /// the exercise base type (from and for backend) normally for the exercise editor
    /// </summary>
    public class ExerciseBase
    {
        public int Id { get; set; }

        /// <summary>
        /// the exercise properties
        /// </summary>
        public ExercisePropertiesFullBase ExerciseProperties { get; set; }

        /// <summary>
        /// the exercise description with the assets
        /// </summary>
        public ExerciseDescriptionFullBase ExerciseDescription { get; set; }

        /// <summary>
        /// the exercise meta data (currently only tags)
        /// </summary>
        public ExerciseMetaDataFullBase MetaData { get; set; }

        /// <summary>
        /// the default test case settings, used when the user creates a custom test
        /// </summary>
        public TestSettingsFullBase DefaultTestSettings { get; set; }

        /// <summary>
        /// the code templates for the plangs
        /// </summary>
        public List<ExerciseCodeTemplateFullBase> CodeTemplates { get; set; }

        /// <summary>
        /// a note field
        /// </summary>
        public string Note { get; set; }

        public ExerciseBase()
        {
            CodeTemplates = new List<ExerciseCodeTemplateFullBase>();
        }
    }

    /// <summary>
    /// a exercise from the frontend for the backend (tests does include file content)
    /// </summary>
    public class ExerciseForBackend : ExerciseBase
    {
        /// <summary>
        /// the tests
        /// </summary>
        public List<ExerciseTestForBackend> Tests { get; set; }

        public ExerciseForBackend()
        {
            Tests = new List<ExerciseTestForBackend>();
        }
    }

    /// <summary>
    /// a exercise from the frontend for the backend (tests does NOT include file content because is cannot changed or viewed in the frontend...)
    /// frontend type is EditorExerciseFromBackend
    /// </summary>
    public class ExerciseFromBackend : ExerciseBase
    {
        /// <summary>
        /// the tests
        /// </summary>
        public List<ExerciseTestFromBackend> Tests { get; set; }

        /// <summary>
        /// the number of releases for the exercise
        /// some features are only available or disabled depending of this value
        /// </summary>
        public int NumReleases { get; set; }

        /// <summary>
        /// the user id of the creator or null (if the owner was deleted, then only the user group can manage the exercise)
        /// </summary>
        public int? CreatorId { get; set; }

        public ExerciseFromBackend()
        {
            Tests = new List<ExerciseTestFromBackend>();
        }
    }

    /// <summary>
    /// exercise like <see cref="ExerciseFromBackend"/> but the tests include the reail asset file data
    /// fromtend type is ExerciseFromBackendWithTestAssetData
    /// </summary>
    public class ExerciseFromBackendWithData : ExerciseBase
    {
        /// <summary>
        /// the tests
        /// </summary>
        public List<ExerciseTestFromBackendWithData> Tests { get; set; }

        /// <summary>
        /// the number of releases for the exercise
        /// some features are only available or disabled depending of this value
        /// </summary>
        public int NumReleases { get; set; }

        /// <summary>
        /// the user id of the creator or null (if the owner was deleted, then only the user group can manage the exercise)
        /// </summary>
        public int? CreatorId { get; set; }

        public ExerciseFromBackendWithData()
        {
            Tests = new List<ExerciseTestFromBackendWithData>();
        }
    }

    /// <summary>
    /// a preview for an asset file
    /// </summary>
    public class FilePreviewFromBackend
    {
        public int Id { get; set; }

        public string OriginalName { get; set; }

        public string MimeType { get; set; }

        public long SizeInBytes { get; set; }
    }

    #endregion

    #region ExerciseTestCase

    /// <summary>
    /// a test base type
    /// </summary>
    public class ExerciseTestBase
    {
        public int Id { get; set; }

        /// <summary>
        /// the display name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// true: is a submit test, false: not (normal test)
        /// </summary>
        public bool IsSubmitTest { get; set; }

        /// <summary>
        /// the test content
        /// </summary>
        public string Content { get; set; }

        /// <summary>
        /// the weight of the test
        /// </summary>
        public int Weight { get; set; }

        /// <summary>
        /// ui proeprty, the display index for the sorting in the frontend
        /// </summary>
        public int DisplayIndex { get; set; }

        /// <summary>
        /// the test type id
        /// </summary>
        public int TestTypeId { get; set; }

        /// <summary>
        /// the test settings
        /// </summary>
        public TestSettingsFullBase TestSettings { get; set; }
    }

    /// <summary>
    /// a test for the backend (assets does contain data)
    /// </summary>
    public class ExerciseTestForBackend : ExerciseTestBase
    {
        /// <summary>
        /// CURRENTLY not used, these are just the previews
        /// but the files (and previews) are managed via separate api calls
        /// </summary>
        public List<FilePreviewFromBackend> Files { get; set; }

        public ExerciseTestForBackend()
        {
            Files = new List<FilePreviewFromBackend>();
        }
    }

    /// <summary>
    /// a test from the frontend (assets does NOT contain data)
    /// </summary>
    public class ExerciseTestFromBackend : ExerciseTestBase
    {
        /// <summary>
        /// the assets (files) (without data because could not viewed or changed at frontend)
        /// </summary>
        public List<FilePreviewFromBackend> Files { get; set; }

        public ExerciseTestFromBackend()
        {
            Files = new List<FilePreviewFromBackend>();
        }
    }

    /// <summary>
    /// a test from the frontend with asset data
    /// </summary>
    public class ExerciseTestFromBackendWithData : ExerciseTestBase
    {
        /// <summary>
        /// the assets (files) (without data because could not viewed or changed at frontend)
        /// </summary>
        public List<TestAssetFromBackendWithData> Files { get; set; }

        public ExerciseTestFromBackendWithData()
        {
            Files = new List<TestAssetFromBackendWithData>();
        }
    }

    #endregion

    #region TestCaseAsset

    /// <summary>
    /// a test asset
    /// </summary>
    public class TestAssetBase
    {
        public int Id { get; set; }

        /// <summary>
        /// the display name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the mime type
        /// </summary>
        public string MimeType { get; set; }

        /// <summary>
        /// the size of the asset (when sending to frontend do not include content)
        /// because could not be changed or viewed
        /// </summary>
        public long Size { get; set; }
    }

    /// <summary>
    /// a full test asset maybe with data (can be null)
    /// </summary>
    public class TestAssetFromBackendWithData : TestAssetBase
    {
        /// <summary>
        /// the content of the asset
        /// CAN BE NULL if file could not be read
        /// </summary>
        public List<byte> Content { get; set; }

        /// <summary>
        /// the hash of the content
        /// </summary>
        public string Hash { get; set; }
    }

    #endregion

    #endregion

    #endregion
}
